#include "ShareFileTransfer.h"
#include "ShareStrings.h"
#include "ShareWindow.h"
#include <app/Roster.h>
#include <kernel/OS.h>
#include <kernel/fs_attr.h>
#include <interface/ListView.h>
#include <interface/Region.h>
#include <storage/Mime.h>
#include <storage/NodeInfo.h>

#include "dataio/TCPSocketDataIO.h"
#include "stringtokenizer/StringTokenizer.h"
#include "iogateway/PortableMessageIOGateway.h"

#include "Colors.h"
#include "ShareUtils.h"
#include "md5.h"

#define MD5_DIGEST_SIZE 16
#define ANIM_RECT_SPACING 16
#define ANIM_RECT_WIDTH   16

const rgb_color PalePink  = {255, 233, 233, 255};
const rgb_color PaleBlue  = {233, 233, 255, 255};
const rgb_color PaleGreen = {233, 255, 233, 255};

const rgb_color BrightPink  = {255, 200, 200, 255};
const rgb_color BrightBlue  = {200, 200, 255, 255};
const rgb_color BrightGreen = {200, 255, 200, 255};

const rgb_color PolePink  = {255, 216, 216, 255};
const rgb_color PoleBlue  = {216, 216, 255, 255};
const rgb_color PoleGreen = {216, 255, 216, 255};

/* Abstract base class */
class MD5Looper : public BLooper 
{
public:
   MD5Looper()
   {
      // empty
   }

   status_t StartProcessing(int32 replyCode, PortableMessageRef msgRef, const BMessenger & replyTo, volatile bool * shutdownFlag)
   {
      _replyCode = replyCode;
      _msgRef    = msgRef;
      _replyTo   = replyTo; 
      _shutdownFlag = shutdownFlag;

      if (Run() >= 0)
      {
         BMessage whatever;
         PostMessage(&whatever);  // just to get us going
         return B_NO_ERROR;
      }
      else return B_ERROR;
   }

   virtual void MessageReceived(BMessage *) = 0;

   void AddEntryRef(const entry_ref & ref) {_entryRefs.AddTail(ref);}
   PortableMessageRef GetMessageRef() {return _msgRef;}

protected:
   void Reply()
   {
      BMessage reportDone(_replyCode);
      reportDone.AddPointer("from", this);
      _replyTo.SendMessage(&reportDone);
   }

   int32 _replyCode;
   PortableQueue<entry_ref> _entryRefs;
   PortableMessageRef _msgRef;
   BMessenger _replyTo;
   volatile bool * _shutdownFlag;
};


/* Background thread used for reading MD5 hashcodes into our TRANSFER_COMMAND_FILE_LIST method, for sending */
class MD5SendLooper : public MD5Looper
{
public:
   virtual void MessageReceived(BMessage *)
   {
      PortableMessage & msg = *_msgRef.GetItemPointer();
      int numRefs = _entryRefs.CountItems();
      for (int i=0; i<numRefs; i++) // (msg.FindString("files", i, &nextStr) == B_NO_ERROR); i++)
      {
         // Add current file length and md5, if any
         off_t fileOffset = 0;  // i.e. autodetect file size for offset
         uint8 digest[MD5_DIGEST_SIZE];
         (void) HashFileMD5(BFile(&_entryRefs[i], B_READ_ONLY), fileOffset, digest, _shutdownFlag);
         msg.AddInt64("offsets", fileOffset);
         msg.AddData("md5", B_RAW_TYPE, digest, (fileOffset > 0) ? sizeof(digest) : 1);
      }
      Reply();
   }
};


/* Background thread used for reading MD5 hashcodes from files to verify them against the codes sent by the remote client */
class MD5ReceiveLooper : public MD5Looper
{
public:
   virtual void MessageReceived(BMessage *)
   {
      PortableMessage * msg = _msgRef.GetItemPointer();
      int numRefs = _entryRefs.CountItems();
      for (int i=0; i<numRefs; i++) 
      {
         // Check to see if the requestor already has the beginning of this file
         // If he does, we can do the old auto-resume-download trick!
         off_t offset = 0LL;
         if ((msg->FindInt64("offsets", i, &offset) == B_NO_ERROR)&&(offset > 0))
         {
            off_t uploadOffset = 0LL;  // default to starting the upload at the beginning of the file
            const uint8 * hisDigest;
            size_t numBytes;
            if ((msg->FindData("md5", B_RAW_TYPE, i, (const void **) &hisDigest, &numBytes) == B_NO_ERROR)&&(numBytes == MD5_DIGEST_SIZE))
            {
               // Okay, the requestor has told us enough so that we can check our local file...
               uint8 myDigest[MD5_DIGEST_SIZE];
               for (size_t h=0; h<ARRAYITEMS(myDigest); h++) myDigest[h] = 'x';  // so I know when it hasn't been overwritten

               BFile tempFile;
               if ((HashFileMD5(BFile(&_entryRefs[i], B_READ_ONLY), offset, myDigest, _shutdownFlag) == B_NO_ERROR)&&
                   (memcmp(hisDigest, myDigest, sizeof(myDigest)) == 0)) uploadOffset = offset;
            }
            msg->ReplaceInt64("offsets", i, uploadOffset);
         }
      }
      Reply();
   }
};

static uint32 MD5LooperPointerHashFunc(MD5Looper * const & ptr);
uint32 MD5LooperPointerHashFunc(MD5Looper * const & ptr)
{
   return (uint32) ptr;
}

ShareFileTransfer ::
ShareFileTransfer(const BDirectory & fileDir, const char * localSessionID) :
	_dir(fileDir),
	_fileSet(PortableStringHashFunc),
	_origFileSet(PortableStringHashFunc),
	_md5Loopers(MD5LooperPointerHashFunc),
	_currentFileStartTime(0),
	_currentFileStartOffset(0),
	_currentFileSize(-1),
	_currentFileBytesDone(-1),
        _autoRestart(false),
	_mtt(NULL),
	_uploadSession(false),
	_onConnect(TRANSFER_COMMAND_CONNECTED_TO_PEER),
	_onDisconnect(TRANSFER_COMMAND_DISCONNECTED_FROM_PEER),
	_onEmpty(TRANSFER_OUTPUT_QUEUE_EMPTY),
	_bitmap(NULL),
	_localSessionID(localSessionID),
	_isAccepting(false),
	_isConnecting(false),
	_isConnected(false),
	_isWaitingOnLocal(true),
	_isWaitingOnRemote(false),
	_wasConnected(false),
	_errorOccurred(false),
	_isFinished(false),
	_currentFileIndex(0),
	_origFileSetSize(0),
	_animShift(0.0f),
	_lastTransferTime(0LL),
	_shutdownFlag(false),
        _beginTransferEnabled(true)
{
   // empty
}

ShareFileTransfer ::
~ShareFileTransfer()
{
   _shutdownFlag = true;  // get all child loopers to stop immediately

   // Clean up any MD5 loopers still in progress
   PortableHashtableIterator<MD5Looper *, bool> iter = _md5Loopers.GetIterator();
   MD5Looper * next;
   while(iter.GetNextKey(next)) if (next->Lock()) next->Quit();

   delete _mtt;
}

void 
ShareFileTransfer::
AddRequestedFileName(const char * fileName, off_t optStartByte)
{
   _origFileSet.Put(PortableString(fileName), optStartByte);
   _fileSet.Put(PortableString(fileName), optStartByte);
}

status_t 
ShareFileTransfer::
InitConnectSession(const char * hostName, int port, const char * remoteSessionID)
{
   _remoteHostName = hostName;
   _remotePort = port;
   _remoteSessionID = remoteSessionID;
   _remoteUserName = ((ShareWindow*)Looper())->GetUserNameForSession(remoteSessionID);
   _origFileSetSize = _origFileSet.Size();
   _fileSet = _origFileSet;
   _fileSetIter = _fileSet.GetIterator();
   _currentFileBytesDone = _currentFileSize = -1;
   _currentFileStartOffset = 0;
   _currentFileStartTime = 0;
   _uploadSession = (_origFileSetSize == 0);
   if (_uploadSession) ((ShareWindow*)Looper())->DoBeep(SYSTEM_SOUND_UPLOAD_STARTED);

   BMessenger toMe(this);
   delete _mtt;  // just in case there is an old one around for some reason
   _mtt = new MessageTransceiverThread(toMe);
   if (_uploadSession) _mtt->SetOutgoingQueueDrainedMessage(&_onEmpty);  // so we get notified when it's time to send more data
   return B_NO_ERROR;
}

status_t
ShareFileTransfer ::
InitAcceptSession(const char * remoteSessionID)
{
   _isAccepting = true;
   _remoteSessionID = remoteSessionID;
   _remoteUserName = ((ShareWindow*)Looper())->GetUserNameForSession(remoteSessionID);
   _origFileSetSize = _origFileSet.Size();
   _fileSet = _origFileSet;
   _fileSetIter = _fileSet.GetIterator();
   _currentFileBytesDone = _currentFileSize = -1;
   _currentFileStartOffset = 0;
   _currentFileStartTime = 0;
   _uploadSession = (_origFileSetSize == 0);
   if (_uploadSession) ((ShareWindow*)Looper())->DoBeep(SYSTEM_SOUND_UPLOAD_STARTED);

   BMessenger toMe(this);
   delete _mtt;  // just in case there is an old one around for some reason
   _mtt = new MessageTransceiverThread(toMe);
   if (_uploadSession) _mtt->SetOutgoingQueueDrainedMessage(&_onEmpty);  // so we get notified when it's time to send more data

   // First try the "standard" ports (to satisfy those firewall-configging guys),
   // if none of those are available, we'll use any old port.
   status_t ret = B_ERROR;
   for (int i=DEFAULT_LISTEN_PORT; i<=DEFAULT_LISTEN_PORT+LISTEN_PORT_RANGE; i++)
   {
      if ((ret = _mtt->StartAcceptThread((i<DEFAULT_LISTEN_PORT+LISTEN_PORT_RANGE)?i:0, false, &_onConnect, &_onDisconnect)) == B_NO_ERROR) break;
   }

   if (ret == B_NO_ERROR) _isAccepting = true;
   UpdateTransferTime();  // enable inactivity timeout when waiting for a connection
   return ret;
}


status_t 
ShareFileTransfer::
InitSocketUploadSession(int socket, bool sendQueuedNotify)
{
   _isConnected = _uploadSession = true;
   ((ShareWindow*)Looper())->DoBeep(SYSTEM_SOUND_UPLOAD_STARTED);
   _origFileSetSize = -1;  // we'll know this when we get the TRANSFER_COMMAND_FILE_LIST message

   BMessenger toMe(this);

   delete _mtt;  // just in case there is an old one around for some reason
   _mtt = new MessageTransceiverThread(toMe);
   _mtt->SetOutgoingQueueDrainedMessage(&_onEmpty);  // so we get notified when it's time to send more data
   status_t ret = _mtt->StartThread(new PortableMessageIOGateway(new TCPSocketDataIO(socket, false)), socket, &_onDisconnect);
   if ((ret == B_NO_ERROR)&&(sendQueuedNotify)) SendNotifyQueuedMessage();
   return ret;
}


void
ShareFileTransfer ::
SendNotifyQueuedMessage()
{
   PortableMessage * notifyMsg = _mtt->GetMessagePool()->GetObject();
   if (notifyMsg)
   {
      PortableMessageRef msgRef(notifyMsg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
      notifyMsg->what = TRANSFER_COMMAND_NOTIFY_QUEUED;
      MASSERT(msgRef.GetItemPointer(), "queuemsg:  no item pointer?");
      _mtt->SendOutgoingMessage(msgRef);
      ResetTransferTime(); // don't timeout if he's queued, in this case it's okay that the connection is idle
   }
}


void 
ShareFileTransfer::
MessageReceived(BMessage * msg)
{
   switch(msg->what)
   {
      case TRANSFER_COMMAND_MD5_RECV_READ_DONE:
      {
         MD5Looper * looper;
         if ((msg->FindPointer("from", (void **) &looper) == B_NO_ERROR)&&(_md5Loopers.Remove(looper)))
         {
            PortableMessageRef msgRef;
            msgRef = looper->GetMessageRef();
            if (looper->Lock()) looper->Quit();

            PortableMessage * pmsg = msgRef.GetItemPointer();
            const char * nextFile;
            for (int i=0; (pmsg->FindString("files", i, &nextFile) == B_NO_ERROR); i++)
            {
               uint64 offset = 0LL;
               (void) pmsg->FindInt64("offsets", i, (int64*) &offset);
               AddRequestedFileName(nextFile, offset);
            }

            int numFiles = _fileSet.Size();
            if (numFiles > 0) 
            {
               _origFileSetSize = numFiles;
               _fileSetIter = _fileSet.GetIterator();
               ((ShareWindow*)Looper())->RefreshTransferItem(this);
               DoUpload();
            }
            else AbortSession(true);   // he's gotta ask for something, otherwise why did he connect?
         }
      }
      break;

      case TRANSFER_COMMAND_MD5_SEND_READ_DONE:
      {
         MD5Looper * looper;
         if ((msg->FindPointer("from", (void **) &looper) == B_NO_ERROR)&&(_md5Loopers.Remove(looper)))
         {
            PortableMessageRef msgRef;
            msgRef = looper->GetMessageRef();
            if (looper->Lock()) looper->Quit();
            if (_mtt) _mtt->SendOutgoingMessage(msgRef);
            ((ShareWindow*)Looper())->RefreshTransferItem(this);
         }
      }
      break;

      case TRANSFER_COMMAND_CONNECTED_TO_PEER:
      {
         _isConnecting = false;
         _isConnected = _wasConnected = true;
         if (_mtt)  // paranoia
         {
            if (_uploadSession)
            {
               if (_isWaitingOnLocal) SendNotifyQueuedMessage();
            }
            else
            {
               // Tell the remote BeShare connection what files we want from him
               PortableMessage * fileRequest = _mtt->GetMessagePool()->GetObject();
               if (fileRequest)
               {
                  PortableMessageRef fileRequestRef(fileRequest, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
                  fileRequest->what = TRANSFER_COMMAND_FILE_LIST;
                  fileRequest->AddString("beshare:FromSession", _localSessionID);
                  PortableHashtableIterator<PortableString, off_t> iter = _fileSet.GetIterator();
                  PortableString * next;

                  MD5Looper * md5Looper = new MD5SendLooper;
                  while((next = iter.GetNextKey()) != NULL)
                  {
                     fileRequest->AddString("files", next->Cstr());
                     BEntry ent(&_dir, next->Cstr(), true);
                     entry_ref er;
                     (void) ent.GetRef(&er);
                     md5Looper->AddEntryRef(er);
                  }

                  // Calculating the md5 hashes could take a long time if the files are big, so I'll do it in a background thread!
                  if (md5Looper->StartProcessing(TRANSFER_COMMAND_MD5_SEND_READ_DONE, fileRequestRef, BMessenger(this), &_shutdownFlag) == B_NO_ERROR) 
                  {
                     _md5Loopers.Put(md5Looper, true);
                     ((ShareWindow*)Looper())->RefreshTransferItem(this);
                  }
                  else 
                  {
                     delete md5Looper;  // ???
                     AbortSession(true);
                  }
               }
               else AbortSession(true);
            }
         }
         ((ShareWindow*)Looper())->FileTransferConnected(this);
      }
      break;

      case TRANSFER_COMMAND_DISCONNECTED_FROM_PEER:
      {
         _shutdownFlag = true;  // no sense reading any more md5 data if we're no longer connected!
         _isFinished = true;
         _isConnecting = _isConnected = _isAccepting = _isWaitingOnLocal = _isWaitingOnRemote = false;
         if ((_currentFileBytesDone < _currentFileSize)||(_currentFileIndex < _origFileSetSize)) _errorOccurred = true;
         else 
         {
            _autoRestart = false; // no need to restart, we're done
            ((ShareWindow*)Looper())->DoBeep(SYSTEM_SOUND_DOWNLOAD_FINISHED);
         }

         if (_autoRestart == false) ((ShareWindow*)Looper())->FileTransferDisconnected(this);
         delete _mtt;  // might as well get rid of it, it's useless now
         _mtt = NULL;
         ResetTransferTime();
   
         if (_autoRestart)
         {
            _autoRestart = false;
            RestartSession();
            ((ShareWindow*)Looper())->DequeueTransferSessions();   
         }
      }
      break;

      case TRANSFER_OUTPUT_QUEUE_EMPTY:
         if (_uploadSession) DoUpload();
      break;

      case PORTABLE_MESSAGES_RECEIVED:
      {
         // This BMessage comes from our MessageTransceiverThread, and
         // tells us that there may be some incoming PortableMessages to look at.
         if (_mtt)
         {
            PortableMessageRef next;
            while((_mtt)&&(_mtt->GetNextIncomingMessage(next) == B_NO_ERROR)) PortableMessageReceived(next);
         }
      }
      break;            
   }
}

uint32
ShareFileTransfer ::
CalculateChecksum(const uint8 * data, size_t bufSize) const
{
   uint32 sum = 0L;
   for (size_t i=0; i<bufSize; i++) sum += (*(data++)<<(i%24));
   return sum;
}

void
ShareFileTransfer ::
DoUpload()
{
   if (_isWaitingOnLocal) return;   // not yet!

   if (_mtt)
   {
      if (_currentFileBytesDone >= 0)
      {
         PortableMessage * msg = _mtt->GetMessagePool()->GetObject();
         const size_t bufferSize = 8 * 1024;  // 8k buffers should minimize overhead but keep a nice GUI update rate
         uint8 * scratchBuffer;
         if ((msg)&&(msg->AddData("data", B_RAW_TYPE, NULL, bufferSize) == B_NO_ERROR)&&
                    (msg->FindDataPointer("data", B_RAW_TYPE, (void **) &scratchBuffer, NULL) == B_NO_ERROR))
         {
            UpdateTransferTime();  // avoid the moribund-connection-reaper
            PortableMessageRef msgRef(msg, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
            ssize_t numBytes = _currentFile.Read(scratchBuffer, bufferSize);
            if (numBytes > 0) 
            {
               msg->what = TRANSFER_COMMAND_FILE_DATA;

               msg->AddInt32("chk", CalculateChecksum(scratchBuffer, numBytes));  // a little paranoioa, due to file-resumes not working.... (TCP should handle this BUT...)

               if ((size_t)numBytes < bufferSize)
               {
                  // oops!  Better chop the extra bytes out of the field!
                  msg->AddData("temp", B_RAW_TYPE, scratchBuffer, numBytes);  // copy just the valid bytes
                  msg->Rename("temp", "data");                                // and replace the old buffer
               }

               MASSERT(msgRef.GetItemPointer(), "datamsg:  no item pointer?");
               _mtt->SendOutgoingMessage(msgRef);
               _currentFileBytesDone += numBytes;
               ((ShareWindow *)Looper())->RefreshTransferItem(this);
               return;  
            }

            if (numBytes == 0)
            {
               // move to the next file
               _currentFileBytesDone = _currentFileSize = -1;
               _currentFileStartTime = 0;
               _currentFileStartOffset = 0;
               DoUpload();
               return;
            }
         }
         else
         {
            WARN_OUT_OF_MEMORY;
            AbortSession(true);
            return;
         }
      }
      else 
      {
         // Get the next file from our send list
         if (_fileSetIter.GetNextKey(_currentFileName))
         {
            _fileSetIter.GetNextValue(_currentFileBytesDone);  // get the starting byte to read from...
            _currentFileStartTime = system_time();
            _currentFileStartOffset = _currentFileBytesDone;

            _currentFileEntry = ((ShareWindow*)Looper())->FindSharedFile(_currentFileName.Cstr());
            if (_currentFile.SetTo(&_currentFileEntry, B_READ_ONLY) == B_NO_ERROR)
            {
               PortableMessage * header = _mtt->GetMessagePool()->GetObject();
               PortableMessageRef headerMsgRef(header, _mtt->GetMessagePool(), _mtt->GetMessageRefPool());
               if ((header)&&
                   (_currentFile.GetSize(&_currentFileSize) == B_NO_ERROR) &&
                   (_currentFile.Seek(_currentFileBytesDone, SEEK_SET) == _currentFileBytesDone))
               {
                  _currentFileIndex++;

                  PortableString strng(str(STR_USER));
                  strng += _remoteSessionID;
                  if ((_remoteUserName.Length() > 0)&&(_remoteUserName[0]))
                  {
                     strng += str(STR_AKA);
                     strng += _remoteUserName;
                     strng += ")";
                  }
                  strng += str(STR_IS_DOWNLOADING);
                  strng += _currentFileName;
                  ((ShareWindow*)Looper())->LogMessage(LOG_UPLOAD_EVENT_MESSAGE, strng.Cstr());

                  header->AddString("beshare:File Name", _currentFileName.Cstr());
                  header->AddInt64("beshare:File Size", _currentFileSize);
                  header->AddString("beshare:FromSession", _localSessionID);
                  if (_currentFileBytesDone > 0LL) header->AddInt64("beshare:StartOffset", _currentFileBytesDone);  // resume-download mode!

                  char attrName[B_ATTR_NAME_LENGTH];
                  BNodeInfo ni(&_currentFile);
                  if ((ni.InitCheck() == B_NO_ERROR)&&(ni.GetType(attrName) == B_NO_ERROR))
                  {
                     header->AddString("beshare:Kind", attrName);
                     _bitmap = ((ShareWindow*)Looper())->GetBitmap(attrName);
                  }
                  else _bitmap = ((ShareWindow*)Looper())->GetBitmap(NULL); 

                  while(_currentFile.GetNextAttrName(attrName) == B_NO_ERROR)
                  {
                     struct attr_info attrInfo;
                     if ((_currentFile.GetAttrInfo(attrName, &attrInfo) == B_NO_ERROR)&&(attrInfo.size > 0))
                     {
                        char * attrData = new char[attrInfo.size];
                        if (_currentFile.ReadAttr(attrName, attrInfo.type, 0L, attrData, attrInfo.size) == attrInfo.size) (void) header->AddData(attrName, attrInfo.type, attrData, attrInfo.size);
                        delete [] attrData;
                     }
                  }

                  ((ShareWindow*)Looper())->RefreshTransferItem(this);

                  header->what = TRANSFER_COMMAND_FILE_HEADER;
                  MASSERT(headerMsgRef.GetItemPointer(), "headerMsgRef: no itempointer");
                  _mtt->SendOutgoingMessage(headerMsgRef);
               }
               DoUpload();
            }
         }
         else AbortSession(false); // All done!
         return;
      }
   }

   // If we got here, there was an error
   AbortSession(true);
}



void 
ShareFileTransfer ::
PortableMessageReceived(PortableMessageRef msgRef)
{
   PortableMessage * msg = msgRef.GetItemPointer();
   UpdateTransferTime();
   switch(msg->what)
   {
      case TRANSFER_COMMAND_NOTIFY_QUEUED:
         if (_uploadSession == false) 
         {
            _isWaitingOnRemote = true;
            ((ShareWindow *)Looper())->RefreshTransferItem(this);
            ResetTransferTime(); // If we're queued, don't timeout, as we may not be getting data for a long time anyway
         }
      break;

      // Received when the remote peer wishes to tell us what files he wants to download from us
      case TRANSFER_COMMAND_FILE_LIST:
      {
         const char * file;
         if ((_uploadSession)&&(msg->FindString("files", &file) == B_NO_ERROR))  // make sure he's asking for at least one file
         {
            PortableString old = _remoteSessionID;

            if (msg->FindString("beshare:FromSession", _remoteSessionID) == B_NO_ERROR) 
            {
               _remoteUserName = ((ShareWindow*)Looper())->GetUserNameForSession(_remoteSessionID.Cstr());
            }
            else _remoteUserName = str(STR_UNKNOWN);

            // in case our session ID allows us to start...
            if (old != _remoteSessionID) ((ShareWindow*)Looper())->DequeueTransferSessions();

            MD5Looper * md5Looper = new MD5ReceiveLooper();
            const char * nextFile;
            for (int i=0; (msg->FindString("files", i, &nextFile) == B_NO_ERROR); i++) md5Looper->AddEntryRef(((ShareWindow*)Looper())->FindSharedFile(nextFile));

            if (md5Looper->StartProcessing(TRANSFER_COMMAND_MD5_RECV_READ_DONE, msgRef, BMessenger(this), &_shutdownFlag) == B_NO_ERROR) 
            {
               _md5Loopers.Put(md5Looper, true);
               ((ShareWindow*)Looper())->RefreshTransferItem(this);
            }
            else 
            {
               delete md5Looper;
               AbortSession(true);
            }
         }
         else AbortSession(true);  // you're going the wrong way buddy!
      }
      break;      

      // When downloading, this indicates the start of the next incoming file
      case TRANSFER_COMMAND_FILE_HEADER:
      {
         PortableString prevFile(_currentFileName);
         PortableString old(_remoteSessionID);
         if ((_uploadSession == false)                                             &&
             (msg->FindString("beshare:File Name", _currentFileName) == B_NO_ERROR)&&
             (msg->FindInt64("beshare:File Size", &_currentFileSize) == B_NO_ERROR)&&
             (msg->FindString("beshare:FromSession", _remoteSessionID) == B_NO_ERROR)&&
             (_fileSet.Remove(_currentFileName)))  // note that while downloading we ignore the value of this hashtable entry
         {
            // so we won't try to get it again on restart...
            (void) _origFileSet.Remove(prevFile);

            _currentFileStartTime = system_time();
            off_t startByte = 0LL;

            bool append = false;
            if ((msg->FindInt64("beshare:StartOffset", &startByte) == B_NO_ERROR)&&(startByte > 0LL))
            {
               // Double check:  Do we have a file named _currentFileName that is exactly the right size?
               // If so, go to append mode; if not, fall back to the old way
               off_t curFileSize;
               if ((_currentFile.SetTo(&_dir, _currentFileName.Cstr(), B_READ_ONLY) == B_NO_ERROR)&&
                   (_currentFile.GetSize(&curFileSize) == B_NO_ERROR)&&(curFileSize == startByte)) append = true;
            }

            if (((append)||(GenerateNewFilename(_currentFileName) == B_NO_ERROR))&&
                (_currentFile.SetTo(&_dir, _currentFileName.Cstr(), B_WRITE_ONLY|B_CREATE_FILE|(append?B_OPEN_AT_END:0)) == B_NO_ERROR))
            {
               BEntry e;
               if (_dir.FindEntry(_currentFileName.Cstr(), &e, true) == B_NO_ERROR) (void) e.GetRef(&_currentFileEntry);
               _isWaitingOnRemote = false;  // if we're getting data, we must not be waiting anymore!
               _remoteUserName = ((ShareWindow*)Looper())->GetUserNameForSession(_remoteSessionID.Cstr());
               _currentFileIndex++;
               _currentFileStartOffset = _currentFileBytesDone = startByte;
               PortableMessageFieldNameIterator iter = msg->GetFieldNameIterator();
               PortableString fieldName;
               while(iter.GetNextFieldName(fieldName))
               {
                  if (strncmp(fieldName.Cstr(), "beshare:", 8))  // don't save our attributes, they aren't necessary
                  {
                     const void * attrData;
                     size_t attrSize;
                     int32 c;
                     type_code type;
                     if ((msg->GetInfo(fieldName.Cstr(), &type, &c)                   == B_NO_ERROR)&&
                         (msg->FindData(fieldName.Cstr(), type, &attrData, &attrSize) == B_NO_ERROR))
                     {
                        (void)_currentFile.WriteAttr(fieldName.Cstr(), type, 0, attrData, attrSize);
                     }
                  }
               }

               BNodeInfo ni(&_currentFile);
               const char * mimeString;
               if ((ni.InitCheck() == B_NO_ERROR)&&(msg->FindString("beshare:Kind", &mimeString) == B_NO_ERROR))
               {
                  ni.SetType(mimeString);
                  _bitmap = ((ShareWindow*)Looper())->GetBitmap(mimeString);
               }
               else _bitmap = ((ShareWindow*)Looper())->GetBitmap(NULL);

               ((ShareWindow*)Looper())->RefreshTransferItem(this);

               // in case our name allows us to start...
               if (old != _remoteSessionID) ((ShareWindow*)Looper())->DequeueTransferSessions();
            }
            else AbortSession(true);
         }
         else AbortSession(true);
      }
      break;

      case TRANSFER_COMMAND_FILE_DATA:
      {
         const uint8 * data;
         size_t numBytes;
         if ((_uploadSession == false)&&(msg->FindData("data", B_RAW_TYPE, (const void **) &data, &numBytes) == B_NO_ERROR))
         {
            bool abortSession = false;

            // paranoia!!!  This shouldn't be necessary, since TCP is supposed to handle this sort of thing.  But I want to check anyway, just in case.
            uint32 checksum;
            if (msg->FindInt32("chk", (int32*)&checksum) == B_NO_ERROR)
            {
               uint32 myChecksum = CalculateChecksum(data, numBytes);  // a little paranoioa (people keep getting munged data -> download-resume failures, why?)
               if (myChecksum != checksum)
               {
                  PortableString errStr("Data Checksum mismatch in file [");
                  errStr += _currentFileName;

                  char temp[256];
                  sprintf(temp, "] (mine=%lu, his=%lu, %lu bytes)", myChecksum, checksum, numBytes);
                  errStr += temp;

                  errStr += " This shouldn't ever happen... looks like something is garbling your incoming TCP stream!?!?";

                  ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, errStr.Cstr());
                  abortSession = true;
               }
            }

            _isWaitingOnRemote = false;  // obviously....
            if (abortSession == false)
            {
               if (_currentFile.Write(data, numBytes) == (ssize_t) numBytes) 
               {
                  _autoRestart = true;  // mark to restart on error only if we got something (to avoid endless cycling)
                  _currentFileBytesDone += numBytes;
                  if (_currentFileBytesDone >= _currentFileSize) _origFileSet.Remove(_currentFileName);
                  ((ShareWindow*)Looper())->RefreshTransferItem(this);
               }
               else 
               {
                  PortableString errStr(str(STR_ERROR_WRITING_TO_FILE));
                  errStr += _currentFileName;
                  ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, errStr.Cstr());
                  abortSession = true;
               }
            }
            if (abortSession) AbortSession(true);
         }
         else ((ShareWindow*)Looper())->LogMessage(LOG_WARNING_MESSAGE, "EMPTY FILE DATA MESSAGE?  Please email jaf@lcsaudio.com and let him know you saw this");
      }
      break;
   }
}

status_t 
ShareFileTransfer :: 
GenerateNewFilename(PortableString & str) const
{
   // don't let BFile's pile up on the stack, they are too expensive
   bool fileExists = false;
   {
      BFile file(&_dir, str.Cstr(), B_READ_ONLY);
      if (file.InitCheck() == B_NO_ERROR) fileExists = true;
   }

   if (fileExists)
   {
      // oops, file exists... gotta generate a new file name and try again!

      // Find last word in name
      const char * lastWord = "";
      StringTokenizer tok(str.Cstr());
      const char * t;
      while((t = tok.GetNextToken()) != NULL) lastWord = t;

      // Check last word to make sure it's just a number
      bool okay = true;
      for (const char * l = lastWord; *l != '\0'; l++)
      {
         if ((*l < '0')||(*l > '9'))
         {
            okay = false;
            break;
         }
      }

      if (okay)
      {
         int idx = str.LastIndexOf(lastWord);
         if (idx >= 0) str = str.Substring(0, idx);

         char newSuffix[100];
         sprintf(newSuffix, "%i", atoi(lastWord)+1);
         str += newSuffix;
      }
      else str += " 2";

      GenerateNewFilename(str);  // go again, in case the new name is also taken!
   }
   return B_NO_ERROR;  
}


void
ShareFileTransfer::
AbortSession(bool error, bool forceNoRestart)
{
   if ((error == false)||(forceNoRestart)) _autoRestart = false;  // don't restart if we are done or the user asked to be stopped...
   _isConnecting = _isAccepting = _isConnected = _isWaitingOnLocal = _isWaitingOnRemote = false;
   _currentFileBytesDone = _currentFileSize = -1;
   _currentFileStartOffset = 0;
   _errorOccurred = error;
   _isFinished = true;

   if (_mtt)
   {
      ((ShareWindow*)Looper())->FileTransferDisconnected(this);
      if (error == false) _mtt->WaitForAllMessagesToBeSent();
      delete _mtt;
      _mtt = NULL;
   }
   ResetTransferTime();

   if (_autoRestart)
   {
      _autoRestart = false;
      RestartSession();
      ((ShareWindow*)Looper())->DequeueTransferSessions();   
   }
}


void 
ShareFileTransfer ::
Update(BView * lv, const BFont * font)
{
   BListItem::Update(lv, font);
   SetHeight(TRANSFER_ITEM_HEIGHT);
   font_height fontAttrs;
   font->GetHeight(&fontAttrs);
   _fontHeight = ceil(fontAttrs.ascent) + ceil(fontAttrs.descent);
}

void 
ShareFileTransfer ::
DrawItem(BView * lv, BRect itemRect, bool /*complete*/)
{
   bool active = ((_isWaitingOnLocal)||(_isWaitingOnRemote)||(_isConnected)||(_isConnecting));
   rgb_color backgroundColor = active ? (_uploadSession ? (_beginTransferEnabled ? PalePink : PaleGreen) : PaleBlue) : Grey;
 
   BRegion Region;
   Region.Include(itemRect);
   lv->ConstrainClippingRegion(&Region);

   BRect colorRect = itemRect;
   colorRect.InsetBy(1.0f, 1.0f);  // avoid flicker by not drawing on the outer layer of pixels 
   PortableString text;
   float percent = (_currentFileSize > 0) ? ((float)_currentFileBytesDone)/_currentFileSize : 0.0f;
   BRect fillRectRight(colorRect);
   if (active) fillRectRight.left = (colorRect.left + colorRect.Width()*percent);
   lv->SetHighColor(backgroundColor);
   lv->FillRect(fillRectRight);

   if (active)
   {
      const rgb_color barColor = _uploadSession ? (_beginTransferEnabled ? BrightPink : BrightGreen) : BrightBlue;
      BRect fillRectLeft(colorRect);
      fillRectLeft.right = fillRectRight.left;
      lv->SetHighColor(barColor);
      lv->FillRect(fillRectLeft);
      if (percent > 0.5f) backgroundColor = barColor;  // do text background from whoever has majority

      // Do little scrolly barber pole animation thingy
      {
         _animShift += 2.0f;
         if (_animShift > ANIM_RECT_SPACING*2.0f) _animShift -= (ANIM_RECT_SPACING*2.0f);

         lv->SetHighColor(_uploadSession ? (_beginTransferEnabled ? PolePink : PoleGreen) : PoleBlue);

         BRegion barberRegion;
         barberRegion.Include(fillRectLeft);
         lv->ConstrainClippingRegion(&barberRegion);
         for (float rl=fillRectRight.right+(ANIM_RECT_SPACING*4)-_animShift; rl >= -ANIM_RECT_SPACING; rl -= ANIM_RECT_SPACING*2.0f)
         {
            const BPoint points[4] = {
               BPoint(rl,                   fillRectLeft.top),
               BPoint(rl-ANIM_RECT_WIDTH,   fillRectLeft.top),
               BPoint(rl-ANIM_RECT_SPACING*4, fillRectLeft.bottom),
               BPoint(ANIM_RECT_WIDTH+(rl-ANIM_RECT_SPACING*4), fillRectLeft.bottom)
            };
            lv->FillPolygon(points, ARRAYITEMS(points));
         }
         lv->ConstrainClippingRegion(&Region);
      }
   }

   if (_errorOccurred) text = str(STR_AN_ERROR_OCCURRED);
   else
   {
           if (_isWaitingOnLocal) text = str(STR_QUEUED_LOCAL_MACHINE_TOO_BUSY);
      else if (_isWaitingOnRemote) text = str(STR_QUEUED_REMOTE_MACHINE_TOO_BUSY);
      else if (_isConnected)
      {
         if (_md5Loopers.Size() > 0) text = str(STR_EXAMINING_FILES);
         else
         {
            text = _uploadSession ? str(STR_SENT) : str(STR_RCVD);

            if (_currentFileSize > 0)
            {
               char doneBuf[64], sizeBuf[64];
               GetByteSizeString(_currentFileBytesDone, doneBuf);
               GetByteSizeString(_currentFileSize, sizeBuf);
               char buf[100];
               sprintf(buf, "%s/%s", doneBuf, sizeBuf);
               text += buf;
            }

            if (_currentFileStartTime > 0)
            {
               bigtime_t elapsedTime = system_time() - _currentFileStartTime;
               if (elapsedTime > 0)
               {
                  // Add performance info, e.g. (43KB/sec, 5:05)
                  text += " (";
     
                  uint64 bytesPerSecond = (((_currentFileBytesDone-_currentFileStartOffset)*1000000LL)/elapsedTime);
                  uint64 bytesToGo      = _currentFileSize-_currentFileBytesDone;
                  uint64 secondsToGo    = (bytesPerSecond > 0) ? (bytesToGo/bytesPerSecond) : 0;

                  char buf[64];
                  GetByteSizeString(bytesPerSecond, buf); 
                  text += buf;
                  text += str(STR_SEC);
              
                  if (secondsToGo >= (60*60)) sprintf(buf, "%Lu:%02Lu:%02Lu", secondsToGo/(60*60), (secondsToGo/60)%60, secondsToGo%60);
                                         else sprintf(buf, "%Lu:%02Lu", secondsToGo/60, secondsToGo%60);
                  text += buf;
                  text += ")";
               }
            }
         }
      }
      else
      {
              if (_isAccepting) text = str(STR_AWAITING_CALLBACK);
         else if (_isConnecting) text = str(STR_CONNECTING);
         else if (_isFinished) 
         {
            if (_wasConnected) text = (_fileSet.Size() == 0) ? str(STR_DOWNLOAD_COMPLETE) : str(STR_DOWNLOAD_ABORTED);
                          else text = str(STR_COULDNT_CONNECT);
         }
         else text = _uploadSession ? str(STR_PREPARING_TO_UPLOAD) : str(STR_PREPARING_TO_DOWNLOAD);
      }
   }

   lv->SetLowColor(backgroundColor);
   lv->SetHighColor(Black);

   char line2[64];
   sprintf(line2, "(%i/%i) %s ", _currentFileIndex, _origFileSetSize, _uploadSession?str(STR_TO):str(STR_FROM));
   PortableString line2Str(line2);
   line2Str += _remoteUserName;
   line2Str += " (";
   line2Str += _remoteSessionID;
   line2Str += ")";
   lv->DrawString(line2Str.Cstr(), BPoint(2.0, colorRect.bottom-(_fontHeight+4.0f)));

   const char * cfn = _currentFileName.Cstr();
   if (cfn[0] == '\0')
   {
      // Get the first name from the set then, if possible
      PortableHashtableIterator<PortableString, off_t> iter = _fileSet.GetIterator();
      PortableString * first = iter.GetNextKey();
      if (first) cfn = first->Cstr();
   }
   if (cfn[0] == '\0')
   {
      // Still nothing?  Get the first name from the orig set then, if possible
      PortableHashtableIterator<PortableString, off_t> iter = _origFileSet.GetIterator();
      PortableString * first = iter.GetNextKey();
      if (first) cfn = first->Cstr();
   }

   lv->DrawString(cfn, BPoint((_bitmap?_bitmap->Bounds().Width()+8.0f:4.0f), colorRect.top+_fontHeight));
   lv->DrawString(text.Cstr(), BPoint(colorRect.left+2.0f,colorRect.bottom-3.0f));
   if (_bitmap)
   {
      lv->SetDrawingMode(B_OP_OVER);
      lv->DrawBitmap(_bitmap, BPoint(colorRect.left + 2.0f, colorRect.top));
   }

   // Do border highlighting
   bool selected = IsSelected();
   lv->SetHighColor(selected ? Black : White);
   lv->StrokeLine(BPoint(itemRect.left, itemRect.top), BPoint(itemRect.right, itemRect.top));
   lv->StrokeLine(BPoint(itemRect.left, itemRect.top), BPoint(itemRect.left, itemRect.bottom));
   lv->SetHighColor(selected ? White : Black);
   lv->StrokeLine(BPoint(itemRect.left, itemRect.bottom), BPoint(itemRect.right, itemRect.bottom));
   lv->StrokeLine(BPoint(itemRect.right, itemRect.top), BPoint(itemRect.right, itemRect.bottom));

   lv->ConstrainClippingRegion(NULL);
}



void
ShareFileTransfer ::
BeginTransfer()
{
   if ((_beginTransferEnabled)&&(_isWaitingOnLocal))
   {
      bool startConnect = false;

      _isWaitingOnLocal = false;
      if (_uploadSession)
      {
         if (_isConnected) 
         {
            if (_fileSet.Size() > 0) DoUpload();
         }
         else startConnect = true;
      }
      else
      {
         if (_isAccepting)
         {
            int port = _mtt ? _mtt->GetPort() : -1;
            if (port >= 0) ((ShareWindow*)Looper())->SendConnectBackRequestMessage(_remoteSessionID.Cstr(), port);
                      else ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, str(STR_ERROR_STARTING_DOWNLOAD_NO_ACCEPT_PORT));
         }
         else startConnect = true;
      }

      if (startConnect)
      {
         if (_mtt)
         {
            status_t ret = _mtt->StartConnectThread(_remoteHostName.Cstr(), _remotePort, &_onConnect, &_onDisconnect);
            if (ret == B_NO_ERROR) _isConnecting = true;
                              else ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, str(STR_ERROR_STARTING_DELAYED_CONNECT));
         }
         else ((ShareWindow*)Looper())->LogMessage(LOG_ERROR_MESSAGE, str(STR_ERROR_STARTING_DELAYED_CONNECT_NO_TRANSCEIVER_THREAD));
      }

      ((ShareWindow *)Looper())->RefreshTransferItem(this);
   }
}


void
ShareFileTransfer ::
OpenCurrentFileFolder()
{
    node_ref tempRef;
    tempRef.device = _currentFileEntry.device;
    tempRef.node = _currentFileEntry.directory;
    BDirectory dir(&tempRef);
    ((ShareWindow*)Looper())->OpenTrackerFolder(dir);
}

status_t
ShareFileTransfer ::
LaunchCurrentItem()
{
    return be_roster->Launch(&_currentFileEntry);
}

void
ShareFileTransfer ::
UpdateTransferTime()
{
   _lastTransferTime = ((_isWaitingOnLocal)||(_isWaitingOnRemote)) ? 0LL : system_time();
}

void
ShareFileTransfer ::
ResetTransferTime()
{
   _lastTransferTime = 0LL;
}


void
ShareFileTransfer ::
RestartSession()
{
   // Close any current session....
   AbortSession(false, true);

   // but it isn't over yet!
   _isFinished = false;
   _isWaitingOnLocal = true;
   _currentFileIndex = 0;
   _shutdownFlag = false;  // so our MD5's will work
   
   // Reset our state to the initial state
   _currentFileName = "";    // gotta reset this!
   _fileSet = _origFileSet;  // store this in case we need to restart
   _origFileSetSize = _fileSet.Size();
   _fileSetIter = _fileSet.GetIterator();  // rewind file set
   if (_remoteHostName.Length() > 0) InitConnectSession(_remoteHostName.Cstr(), _remotePort, _remoteSessionID.Cstr());
                                else InitAcceptSession(_remoteSessionID.Cstr());
}

void
ShareFileTransfer ::
RequeueTransfer()
{
   _isWaitingOnLocal = true;
   SendNotifyQueuedMessage();
}
